/*-
 * Copyright (c) 2003-2007 Tim Kientzle
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "test.h"
__FBSDID("$FreeBSD$");

#include <errno.h>
#include <stdlib.h>
#include <string.h>

static size_t nullsize;
static void *nulls;

struct memblock {
	struct memblock *next;
	size_t	size;
	void *buff;
};


struct memdata {
	void *buff;
	struct memblock *first;
	struct memblock *last;
};

#if ARCHIVE_API_VERSION < 2
//static ssize_t	memory_read_skip(struct archive *, void *, size_t request);
#else
//static off_t	memory_read_skip(struct archive *, void *, off_t request);
#endif
static ssize_t	memory_read(struct archive *, void *, const void **buff);
static ssize_t	memory_write(struct archive *, void *, const void *, size_t);


static ssize_t
memory_write(struct archive *a, void *_private, const void *buff, size_t size)
{
	struct memdata *private = _private;
	struct memblock *block;
	const char *p;
	ssize_t n;
	size_t sum = 0;

	(void)a;

	block = (struct memblock *)malloc(sizeof(*block));
	block->size = size;

	/*
	 * TODO: The following assumes zero-copy behavior, which
	 * isn't actually true.  Change write_compression_none to
	 * use zero-copy and see how much faster it is.
	 */
/*
	if ((const char *)nulls <= (const char *)buff
	    && (const char *)buff < (const char *)nulls + nullsize) {
*/
	for (n = size, p = buff; n > 0; n -= 512, p += 512)
		sum += *p;
	if (sum == 0) {
		/* We don't need to store a NULL block. */
		block->buff = NULL;
	} else {
		/* It's a non-NULL block, copy and save it. */
		block->buff = malloc(size);
		memcpy(block->buff, buff, size);
	}

	if (private->last == NULL) {
		private->first = private->last = block;
	} else {
		private->last->next = block;
		private->last = block;
	}
	block->next = NULL;
	return (size);
}

static ssize_t
memory_read(struct archive *a, void *_private, const void **buff)
{
	struct memdata *private = _private;
	struct memblock *block;
	ssize_t size;

	(void)a;

	free(private->buff);
	if (private->first == NULL) {
		private->last = NULL;
		return (ARCHIVE_EOF);
	}

	block = private->first;
	private->first = block->next;

	size = block->size;
	if (block->buff != NULL) {
		private->buff = block->buff;
		*buff = block->buff;
	} else {
		private->buff = NULL;
		*buff = nulls;
	}
	free(block);
	return (size);
}

/*
#if ARCHIVE_API_VERSION < 2
static ssize_t
memory_read_skip(struct archive *a, void *private, size_t skip)
#else
static off_t
memory_read_skip(struct archive *a, void *private, off_t skip)
#endif
{
	(void)a;
	(void)private;
	return (skip);
}
*/

#define KB 1024
#define MB 1024 * KB
#define GB 1024 * MB

DEFINE_TEST(test_tar_large)
{
	struct memdata memdata;
	struct archive_entry *ae;
	struct archive *a;
	off_t  filesize, writesize;

	nullsize = 512 * KB;
	nulls = malloc(nullsize);
	memset(nulls, 0, nullsize);
	memset(&memdata, 0, sizeof(memdata));

	/*
	 * Open an archive for writing.
	 */
	a = archive_write_new();
	archive_write_set_format_pax_restricted(a);
	archive_write_set_bytes_per_block(a, 0); /* No buffering. */
	archive_write_open(a, &memdata, NULL, memory_write, NULL);

	/*
	 * Write a file to it.
	 */
	assert((ae = archive_entry_new()) != NULL);
	archive_entry_copy_pathname(ae, "file");
	archive_entry_set_mode(ae, S_IFREG | 0755);
	filesize = 16;
	filesize *= GB;
	archive_entry_set_size(ae, filesize);

	assertA(0 == archive_write_header(a, ae));
	archive_entry_free(ae);

	while (filesize > 0) {
		writesize = nullsize;
		if (writesize > filesize)
			writesize = filesize;
		assertA(writesize == archive_write_data(a, nulls, writesize));
		filesize -= writesize;
	}

	assert((ae = archive_entry_new()) != NULL);
	archive_entry_copy_pathname(ae, "file2");
	archive_entry_set_mode(ae, S_IFREG | 0755);
	assertA(0 == archive_write_header(a, ae));
	archive_entry_free(ae);


	/* Close out the archive. */
	assertA(0 == archive_write_close(a));
#if ARCHIVE_API_VERSION > 1
	assertA(0 == archive_write_finish(a));
#else
	archive_write_finish(a);
#endif

	/*
	 * Open the same archive for reading.
	 */
	a = archive_read_new();
	archive_read_support_format_tar(a);
	archive_read_open2(a, &memdata, NULL,
	    memory_read, NULL /* memory_read_skip */, NULL);

	/*
	 * Read entries back.
	 */
	assertEqualIntA(a, 0, archive_read_next_header(a, &ae));
	assertEqualString("file", archive_entry_pathname(ae));

	assertEqualIntA(a, 0, archive_read_next_header(a, &ae));
	assertEqualString("file2", archive_entry_pathname(ae));

	assertEqualIntA(a, ARCHIVE_EOF, archive_read_next_header(a, &ae));

	/* Close out the archive. */
	assertA(0 == archive_read_close(a));
#if ARCHIVE_API_VERSION > 1
	assertA(0 == archive_read_finish(a));
#else
	archive_read_finish(a);
#endif

	free(nulls);
}
